/* Title:       JUMP
  
   Author:      Kent Covert (derived from JUMP.PAS by
                    Jonathan Ridler
                    Information Technology Services,
                    The University of Melbourne,
                    Parkville, Victoria,
                    AUSTRALIA, 3052.)
  
   System:      OSU WWW Server
  
   Description: This program will change the current process to appear as
                the user passed as the first parameter.

   Comments:    1) This program requires: CMKRNL, CMEXEC, SYSNAM, and SYSPRV to run.
                2) ***  ALL AUTHORIZED and CURRENT privileges of the original
                        process remain when this program exits ***
  
   Version:
        1.0     12/28/1995 KAC  Initial program version
*/

#include stdio
#include string
#include ctype
#include stdlib
#include starlet
#include descrip
#include ssdef
#include rmsdef
#include uaidef
#include lnmdef
#include psldef
#include uicdef
#include ossdef
#include jpidef
#include prvdef
#include lib$routines
#include "jump.h"

extern void *CTL$GL_PCB;
extern char *PCB$L_JIB, *JIB$T_USERNAME;
extern int *PCB$L_UIC;
extern char CTL$T_USERNAME;
char **jibPtr = NULL;
char *jibUsernamePtr = NULL;
union uicdef uic, *uicPtr;
char cusername[13];
char jusername[13];
char newUsername[13];
static char jobTable[32];
static char defDev[33];
static char defDir[65];
struct dsc$descriptor_s lnmjobTableDesc = { 0, DSC$K_DTYPE_T, DSC$K_CLASS_S, jobTable};

struct itmlst {
  unsigned short bufferLength;
  unsigned short itemCode;
  void *bufferAddress;
  void *returnLengthAddress;
};

void errorExit(char *statusLine, char *errorMessage) {
  printf("Content-Type: text/html\n\n");
  printf("<PRE>\n");
  printf("%s\n",statusLine);
  printf("%s\n",errorMessage);
  printf("</PRE>\n");
  sys$delprc(0,0);
  exit(4);
}

void enablePrivs() {
  unsigned int result;
  unsigned int privs[2];

  privs[0]=PRV$M_CMKRNL+PRV$M_CMEXEC+PRV$M_SYSNAM+PRV$M_SYSPRV;
  privs[1]=0;
  if ((result=sys$setprv(1,privs,0,0)) != SS$_NORMAL) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while setting privs: %x",result);
    errorExit("404 error in setprv",errorMessage);
  }
}

unsigned int execGetUserPtr(void) {
  jibUsernamePtr = *jibPtr;
  return 0;
}

unsigned int execGetUsername(void) {

  strncpy(cusername,&CTL$T_USERNAME,12);
  strncpy(jusername,jibUsernamePtr,12);
  return 0;
}

unsigned int kernelSetUsername(void) {
  strncpy(&CTL$T_USERNAME,newUsername,12);
  strncpy(jibUsernamePtr,newUsername,12);
  return 0;
}

unsigned int kernelSetUIC(void) {
  *uicPtr = uic;
  return 0;
}

int checkAccess(char *newUsername, int argc, char *argv[]) {
  unsigned int result;
  char *i;
  $DESCRIPTOR(lnmprocessDesc,"LNM$PROCESS");
  $DESCRIPTOR(jumpOriginalUsernameDesc,"JUMP_ORIGINAL_USERNAME");
  static char originalUsername[13];
  static unsigned short originalUsernameLength;
  static unsigned int jpiMode;
  struct itmlst trnlnmItmlst[2]       = {{12, LNM$_STRING    , originalUsername , &originalUsernameLength},
                                         { 0, 0              , 0                , 0}};
  struct itmlst getjpiItmlst[2]       = {{ 4, JPI$_MODE      , &jpiMode         , 0},
                                         { 0, 0              , 0                , 0}};

  if ((result=sys$getjpiw(0,0,0,getjpiItmlst,0,0,0)) != SS$_NORMAL) {
    char errorMessage[256];
    sprintf(errorMessage,"Error getting mode information about process: %x",result);
    errorExit("404 Error getting mode",errorMessage);
  }
/*  if (jpiMode != JPI$K_NETWORK) {
    errorExit("404 Not network mode","JUMP can only be run in network mode!");
  } */
  result=sys$trnlnm(0, &lnmprocessDesc, &jumpOriginalUsernameDesc, &PSL$C_EXEC, trnlnmItmlst);
  if (result == SS$_NORMAL) {
    originalUsername[originalUsernameLength]='\0';       /* We're returning to our original username  */
    sprintf(newUsername,"%-12.12s",WWWUSERNAME);         /* Most security for this part is handled by */
    for(i=newUsername;*i;i++) *i=toupper(*i);            /* by the cleanup program.                   */
    return 2;
  } else {
    if (result == SS$_NOLOGNAM) {                        /* We're switching to a new username */
      if (strcmp(getenv("USER"),WWWUSERNAME) && strcmp(getenv("USER"),"KACOVERT")) {
        errorExit("404 Not WWW","JUMP can only be run from the WWW username.");
      }
      if (argc < 2) {
        errorExit("404 No username","No username specified.");
      }
      if (argc > 2) {
        errorExit("404 Too many params","Too many parameters specified to JUMP.");
      }
      sprintf(newUsername,"%-12.12s",argv[1]);
      for(i=newUsername;*i;i++) *i=toupper(*i);
      return 1;
    } else {
      char errorMessage[256];
      sprintf(errorMessage,"Error while getting original username: %x",result);
      errorExit("404 error in trnlnm",errorMessage);
    }
  }
}

void storeOriginalUsername() {
  unsigned int result;
  $DESCRIPTOR(lnmprocessDesc,"LNM$PROCESS");
  $DESCRIPTOR(jumpOriginalUsernameDesc,"JUMP_ORIGINAL_USERNAME");
  static char originalUsername[13];
  struct itmlst crelnmItmlst[2] = {{ 0, LNM$_STRING    , originalUsername, 0},
                                   { 0, 0              , 0               , 0}};
  
  strcpy(originalUsername,getenv("USER"));
  crelnmItmlst[0].bufferLength = strlen(originalUsername);
  if ((result=sys$crelnm(0, &lnmprocessDesc, &jumpOriginalUsernameDesc, &PSL$C_EXEC, crelnmItmlst)) != SS$_NORMAL &&
       result != SS$_SUPERSEDE) {
    char errorMessage[256];
    sprintf(errorMessage,"Error storing original username!: %x",result);
    errorExit("404 Error in crelnm",errorMessage);
  }
}

void setUsername() {
  unsigned int result;
  if (result=sys$cmkrnl(kernelSetUsername,0)) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while setting username: %x",result);
    errorExit("404 error in cmkrnl",errorMessage);
  }
}

void setUIC(void) {
  unsigned int result;
  if (result=sys$cmkrnl(kernelSetUIC,0)) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while setting UIC: %x",result);
    errorExit("404 error in cmkrnl",errorMessage);
  }
}

void setDirectory() {
  unsigned int result;
  struct dsc$descriptor_s defDirDesc  = { 0, DSC$K_DTYPE_T, DSC$K_CLASS_S, defDir+1};
  $DESCRIPTOR(sysDiskDesc, "SYS$DISK");
  $DESCRIPTOR(lnmprocessDesc, "LNM$PROCESS");
  struct itmlst crelnmItmlst[2] = {{ 0, LNM$_STRING    , defDev+1       , 0},
                                   { 0, 0              , 0              , 0}};

  defDirDesc.dsc$w_length = (unsigned short)defDir[0];
  if ((result=sys$setddir(&defDirDesc,0,0)) != RMS$_NORMAL) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while changing directory: %x",result);
    errorExit("404 error in setddir",errorMessage);
  }
  crelnmItmlst[0].bufferLength = (unsigned short) defDev[0];
  if ((result=sys$crelnm(0, &lnmprocessDesc, &sysDiskDesc, &PSL$C_SUPER, crelnmItmlst)) != SS$_NORMAL && result != SS$_SUPERSEDE) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while setting sys$disk: %x %10.10s",result,defDev+1);
    errorExit("404 error in crelnm",errorMessage);
  }
}

void setLNMTables() {
  unsigned int result;
  static unsigned int attributes = LNM$M_TERMINAL;
  static char groupTable[32];
  static unsigned int jobTableLength;
  $DESCRIPTOR(lnmprocessDirectoryDesc, "LNM$PROCESS_DIRECTORY");
  $DESCRIPTOR(lnmgroupDesc, "LNM$GROUP");
  $DESCRIPTOR(lnmjobDesc, "LNM$JOB");
  $DESCRIPTOR(clsnamDesc, "LOGICAL_NAME_TABLE");
  struct itmlst crelnmItmlst[3]       = {{ 4, LNM$_ATTRIBUTES, &attributes    , 0},
                                         {16, LNM$_STRING    , &groupTable    , 0},
                                         { 0, 0              , 0              , 0}};
  struct itmlst trnlnmItmlst[2]       = {{31, LNM$_STRING    , jobTable       , &jobTableLength},
                                         { 0, 0              , 0              , 0}};
  struct itmlst set_securityItmlst[2] = {{ 4, OSS$_OWNER     , &uic           , 0},
                                         { 0, 0              , 0              , 0}};

  /* Point the LNM$GROUP logical to the group table of the new user */
  sprintf(groupTable,"LNM$GROUP_%6.6ho",uic.uic$v_group);
  if ((result=sys$crelnm(0, &lnmprocessDirectoryDesc, &lnmgroupDesc, &PSL$C_KERNEL, crelnmItmlst)) != SS$_NORMAL &&
            result != SS$_SUPERSEDE) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while changing user of LNM$GROUP: %x",result);
    errorExit("404 error in crelnm",errorMessage);
  }

  /* Change the owner of the LNM$JOB table to this new user */
  if ((result=sys$trnlnm(0, &lnmprocessDirectoryDesc, &lnmjobDesc, &PSL$C_KERNEL, trnlnmItmlst)) != SS$_NORMAL) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while getting LNM$JOB name: %x",result);
    errorExit("404 error in trnlnm",errorMessage);
  }

  lnmjobTableDesc.dsc$w_length = jobTableLength;
  if ((sys$set_security(&clsnamDesc, &lnmjobTableDesc, 0, 0, set_securityItmlst, 0, 0)) != SS$_NORMAL) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while changing owner of LNM$JOB table: %x",result);
    errorExit("404 error in set_security",errorMessage);
  }
}

void setLogicals() {
  unsigned int result;
  static char loginDir[256];
  $DESCRIPTOR(sysloginDesc, "SYS$LOGIN");
  $DESCRIPTOR(sysloginDeviceDesc, "SYS$LOGIN_DEVICE");
  $DESCRIPTOR(sysscratchDesc, "SYS$SCRATCH");
  struct itmlst crelnmItmlst1[2] = {{ 0, LNM$_STRING    , loginDir       , 0},
                                    { 0, 0              , 0              , 0}};
  struct itmlst crelnmItmlst2[2] = {{ 0, LNM$_STRING    , defDev+1       , 0},
                                    { 0, 0              , 0              , 0}};

  /* Change definitions of SYS$LOGIN, SYS$LOGIN_DEVICE, and SYS$SCRATCH */
  sprintf(loginDir, "%.*s%.*s",defDev[0],defDev+1,defDir[0],defDir+1);
  crelnmItmlst1[0].bufferLength = strlen(loginDir);
  if ((result=sys$crelnm(0, &lnmjobTableDesc, &sysloginDesc, &PSL$C_EXEC, crelnmItmlst1)) != SS$_NORMAL
            && result != SS$_SUPERSEDE) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while setting SYS$LOGIN: %x",result);
    errorExit("404 error in crelnm",errorMessage);
  }

  crelnmItmlst2[0].bufferLength = strlen(defDev+1);
  if ((result=sys$crelnm(0, &lnmjobTableDesc, &sysloginDeviceDesc, &PSL$C_EXEC, crelnmItmlst2)) != SS$_NORMAL
            && result != SS$_SUPERSEDE) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while setting SYS$LOGIN_DEVICE: %x",result);
    errorExit("404 error in crelnm",errorMessage);
  }

  if ((result=sys$crelnm(0, &lnmjobTableDesc, &sysscratchDesc, &PSL$C_EXEC, crelnmItmlst1)) != SS$_NORMAL
            && result != SS$_SUPERSEDE) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while setting SYS$SCRATCH: %x",result);
    errorExit("404 error in crelnm",errorMessage);
  }
}

void runCleanup() {
  unsigned int result;
  $DESCRIPTOR(lnmsystemDesc,"LNM$SYSTEM");
  $DESCRIPTOR(www_rootDesc,"WWW_ROOT");
  static char www_root[256], cleanupCommand[256];
  static unsigned short www_rootLength;
  $DESCRIPTOR(lnmprocessDesc,"LNM$PROCESS");
  $DESCRIPTOR(jumpOriginalUsernameDesc,"JUMP_ORIGINAL_USERNAME");
  struct itmlst trnlnmItmlst[2]       = {{256, LNM$_STRING    , www_root      , &www_rootLength},
                                         { 0, 0              , 0              , 0}};
  struct dsc$descriptor_s cleanupCommandDesc = {0 , DSC$K_DTYPE_T, DSC$K_CLASS_S, cleanupCommand};

  if ((result=sys$dellnm(&lnmprocessDesc,&jumpOriginalUsernameDesc,&PSL$C_EXEC)) != SS$_NORMAL) {
    char errorMessage[256];
    sprintf(errorMessage,"Error deleting JUMP_ORIGINAL_USERNAME: %x",result);
    errorExit("404 Error in dellnm",errorMessage);
  }
  if ((result=sys$trnlnm(0, &lnmsystemDesc, &www_rootDesc, &PSL$C_EXEC, trnlnmItmlst)) != SS$_NORMAL) {
    errorExit("404 no WWW_ROOT","WWW_ROOT not defined!");
  }
  www_root[www_rootLength-1]='\0';       /* get rid of ending ] */
  sprintf(cleanupCommand,"@%s%s]USEREXEC_CLEANUP.COM",www_root,WWWBINDIR);
  cleanupCommandDesc.dsc$w_length = strlen(cleanupCommand);
  result = lib$do_command(&cleanupCommandDesc);
  printf("Error executing do_command: %x",result);
  sys$delprc(0,0);
}

main(int argc, char *argv[]) {
  int result;
  int mode;  /* 1 to switch to newUsername, 2 to return to old username */
  struct dsc$descriptor_s newUsernameDesc          = {12, DSC$K_DTYPE_T, DSC$K_CLASS_S, newUsername};
  struct itmlst getuaiItmlst[4]       = {{ 4, UAI$_UIC       , &uic           , 0},
                                         {32, UAI$_DEFDEV    , defDev         , 0},
                                         {64, UAI$_DEFDIR    , defDir         , 0},
                                         { 0, 0              , 0              , 0}};

  enablePrivs();

  /* Do nasty Exec and Kernel stuff... */
  jibPtr = (char **)CTL$GL_PCB + (unsigned int)&PCB$L_JIB/sizeof(char *);
  if (result=sys$cmexec(execGetUserPtr,0)) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while setting jibPtr: %x",result);
    errorExit("404 error in cmexec",errorMessage);
  }

  jibUsernamePtr = jibUsernamePtr + (unsigned int)&JIB$T_USERNAME;
  if (result=sys$cmexec(execGetUsername,0)) {
    char errorMessage[256];
    sprintf(errorMessage,"Error while getting username: %x",result);
    errorExit("404 error in cmexec",errorMessage);
  }

  uicPtr = (union uicdef *)CTL$GL_PCB + (unsigned int)&PCB$L_UIC/sizeof(union uicdef *);

  /* Check access and program validity. */
  mode = checkAccess(newUsername,argc,argv);

  if ((result=sys$getuai(0,0,&newUsernameDesc,getuaiItmlst,0,0,0))!=SS$_NORMAL) {
    if (result == RMS$_RNF) {
      char errorMessage[256];
      sprintf(errorMessage,"Username: %s does not exist.",newUsername);
      errorExit("404 bad username",errorMessage);
    }
  }

  storeOriginalUsername();
  setUsername();
  setUIC();
  setLNMTables();
  setDirectory();
  setLogicals();

  if (mode == 2) runCleanup();
}
